/* bwMonitor.c - rate based bwMonitor */

/*
modification history
--------------------
01b,17nov09,suh fixed spr15980 weighted load-balancing issues
01a,28sep08,sam created
*/

/*
DESCRIPTION
This module uses the kernel rate estimator for rate determination. The rate
estimators must have been installed in the packet path using the netfilter
target RATEEST.

This module implements a rate based next hop selection algorithm for multiapth
routing. This selection is based on the initial weights given by the user and
the current traffic load on each of its nexthop links in a multipath route.

The NXTHOP_LOAD_TOLERANCE is the % tolerance of the load-balancer. This defaults
to 10.
*/

#include "linux/module.h"
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/netdevice.h>
#include <linux/gen_stats.h>
#include <net/route.h>
#include <net/ip_fib.h>
#include  <net/netfilter/xt_rateest.h>

#define NXTHOP_LOAD_TOLERANCE   5

/* link states */
#define NXTHOP_LINK_OVERUSED     1
#define NXTHOP_LINK_UNDERUSED    2
#define NXTHOP_LINK_NORMALUSED   3
#define NXTHOP_LINK_MAX          4

#undef BWMON_DEBUG
#ifdef BWMON_DEBUG
#define BWMON_PRINTK(a...)     do { if (debug) printk(KERN_DEBUG a); } while (0)
#else
#define BWMON_PRINTK(a...)
#endif

extern int (*bwmon_multipath_weight_check) (const struct flowi *flp,
                                             struct fib_result *res);

static int bwMonRateInterval = 2;
struct timer_list bwRateTimer;
static int lastHopSelected=0;
static int multipathSelect = 1;
static int weightedHopPrefer = 0;
static int loadTolerance = NXTHOP_LOAD_TOLERANCE;
#ifdef CONFIG_SYSCTL
static struct ctl_table_header * pBwMonitorSysctlHeader;
#endif /* CONFIG_SYSCTL */
#ifdef BWMON_DEBUG
static int debug = 0;

/*******************************************************************************
*
* bwMonTimerHandler - timer handler for bw monitor.
*
* This function calculates the data rate of all the interfaces registered.
*
* RETURNS: 0
*/

void  bwMonTimerHandler
    (
    unsigned long  arg
    )
    {
    struct net_device * pNetDevice=NULL;
    struct net_device_stats * pNetStats=NULL;
    unsigned long tx_rate =0;
    unsigned long rx_rate = 0;
    char rxName[16]={0};
    char txName[16]={0};

    /* loop through the list of devices and calculate the rate of each one */
	for (pNetDevice = dev_base; pNetDevice; pNetDevice = pNetDevice->next)
        {
        struct xt_rateest *est1=NULL;

        tx_rate = 0;
        rx_rate = 0;

        /* continue if the interface is down */
        if ((pNetDevice->flags & IFF_UP) == 0)
            continue;

        /* if the interface does not have statistics counter, then continue */
	    if (pNetDevice->get_stats == NULL)
            continue;

		pNetStats = pNetDevice->get_stats(pNetDevice);
        if (pNetStats == NULL)
            continue;

        snprintf(txName, sizeof(txName), "tx_%s", pNetDevice->name);
        est1 = xt_rateest_lookup (txName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            tx_rate = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        snprintf(rxName, sizeof(rxName), "rx_%s", pNetDevice->name);
        est1 = xt_rateest_lookup (rxName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            rx_rate = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        BWMON_PRINTK ("RATEEST: if:%s rx:%lu tx:%lu total: %lu\n",
                      pNetDevice->name, rx_rate, tx_rate, tx_rate + rx_rate);

        }

    /* reset the timer */
    del_timer(&bwRateTimer);
    bwRateTimer.expires = jiffies + bwMonRateInterval * HZ;
	add_timer(&bwRateTimer);

    return;
    }
#endif /* BWMON_DEBUG */

/*******************************************************************************
*
* bwMonMultipathNxtHopSelect - selects the next-hop for a multipath route.
*
* This function selects the nexthop based on the current linkusage.
*
* RETURNS: 0
*/

int bwMonMultipathNxtHopSelect
    (
    const struct flowi * pFlow,       /* traffic description */
    struct fib_result *  pFib         /* multipath route */
    )
    {
    struct fib_nh * pNxtHop=NULL;
    int    totalBw=0;
    int    netIdealWeight=0;
    int    currHop=0;
    int    diff=0;
    int    weight=0;
    unsigned long  timestamp=0;
    int    selHop=0;
    unsigned long tx_bps = 0;
    unsigned long rx_bps = 0;
    char rxName[16]={0};
    char txName[16]={0};
    int    underusedHopAvailable = 0;

    BWMON_PRINTK ("bwMonMultipathNxtHopSelect: checking rates \n");

    /*
     * 1. reset the link usage of each hop.
     * 2. calculate the total bandwidth routed using all the hops of this route
     * 3. calculate the sum of weights of all hops.
     */
    for (currHop=0; currHop < pFib->fi->fib_nhs; currHop++)
        {
        struct xt_rateest *est1=NULL;

        memset (txName, '\0', sizeof(txName));
        memset (rxName, '\0', sizeof(rxName));

        pNxtHop = &pFib->fi->fib_nh[currHop];

        pNxtHop->linkUsage = NXTHOP_LINK_NORMALUSED;

        snprintf(txName, sizeof(txName), "tx_%s", pNxtHop->nh_dev->name);
        est1 = xt_rateest_lookup (txName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            tx_bps = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        snprintf(rxName, sizeof(rxName), "rx_%s", pNxtHop->nh_dev->name);
        est1 = xt_rateest_lookup (rxName);
        if (est1 != NULL)
            {
            spin_lock_bh(&est1->lock);
            rx_bps = est1->rstats.bps;
            spin_unlock_bh(&est1->lock);
            xt_rateest_put(est1);
            }

        pNxtHop->currBwShare = tx_bps + rx_bps;

        totalBw += pNxtHop->currBwShare;
        netIdealWeight += pNxtHop->nh_weight;

        BWMON_PRINTK ("hop:%d dev:%s TS:%llu weight:%d currentBW:%d bytes/sec\n",
                      currHop, pNxtHop->nh_dev->name, pNxtHop->nh_timestamp,
                      pNxtHop->nh_weight, pNxtHop->currBwShare);
	    }

    if (!totalBw || !netIdealWeight)
        {
        BWMON_PRINTK ("totalBw:%d, netIdealWeight:%d\n", totalBw,
                       netIdealWeight);
        goto nxthopalgo;
        }

    for (currHop=0; currHop < pFib->fi->fib_nhs; currHop++)
        {
        pNxtHop = &pFib->fi->fib_nh[currHop];

        /* 1. calculate the current %bandwidth of each hop */
	    pNxtHop->currBwShare =  (pNxtHop->currBwShare * 100L) / totalBw;

        /* 2. calculate the ideal %bandwidth of each hop */
        if (pNxtHop->idealBwShare == 0)
            {
            pNxtHop->idealBwShare = (pNxtHop->nh_weight * 100L) / netIdealWeight;
            }
	
        BWMON_PRINTK ("current BW share of hop %d:%d ideal:%d\n",
	                 currHop, pNxtHop->currBwShare, pNxtHop->idealBwShare);

        /* 3. Mark if the nexthop is being normally/over/under used */
        diff = pNxtHop->idealBwShare - pNxtHop->currBwShare;

        if (((diff > 0) && (diff <= loadTolerance)) ||
            ((diff < 0) && (-diff <= loadTolerance)))
            {
            BWMON_PRINTK ("hop %d being used normally\n", currHop);
            pNxtHop->linkUsage = NXTHOP_LINK_NORMALUSED;
            }
        else if ((diff > 0) && (diff > loadTolerance))
            {
            BWMON_PRINTK ("hop %d being under used\n", currHop);
            pNxtHop->linkUsage = NXTHOP_LINK_UNDERUSED;

            if (!(pNxtHop->nh_flags&RTNH_F_DEAD))
                underusedHopAvailable = 1;
            }
        else if ((diff < 0) && (-diff > loadTolerance))
            {
            BWMON_PRINTK ("hop %d being over used\n", currHop);
            pNxtHop->linkUsage = NXTHOP_LINK_OVERUSED;
            }
	    }

    nxthopalgo:
    /*
     * The status of the links have been marked. if there were any under-used
     * link, we would have already returned.
     *
     * next hop selection algo:
     *  1. Don't use any over-used hops.
     *  2. If under-used hops are available, select the best hop among the
     *     under-used hops.
     *  3. If under-used hops are not available, select the best hop among the
     *     normal-used hops.
     *
     *  best hop selection algo:
     *  1. If weightedHopPrefer is set, select the hop with the greatest
     *     weight. Resolve ties in favour of the least recently used hop.
     *  2. If weightedHopPrefer is not set, select the least recently used hop.
     */
    selHop  = 0;
    weight = 0;

    for (currHop=0; currHop < pFib->fi->fib_nhs; currHop++)
        {
        pNxtHop = &pFib->fi->fib_nh[currHop];

        /* skip dead links */
        if (pNxtHop->nh_flags&RTNH_F_DEAD)
            continue;

        /* skip over-used links */
        if (pNxtHop->linkUsage == NXTHOP_LINK_OVERUSED)
            continue;

        /* skip normal-used links when under-used links exist */
        if (underusedHopAvailable
            && pNxtHop->linkUsage == NXTHOP_LINK_NORMALUSED)
            continue;

        if (!weight)
            {
            selHop  = currHop;
            weight = pNxtHop->nh_weight;
            timestamp = pNxtHop->nh_timestamp;
            /* This is the first hop that we looked into */
            continue;
            }

        /* If weightedHopPrefer is set, prefer the hop with the greatest
         * weight, resolving ties in favour of the LRU hop.
         */
        if (weightedHopPrefer
            && pNxtHop->nh_weight > weight)
            {
            selHop = currHop;
            timestamp = pNxtHop->nh_timestamp;
            weight = pNxtHop->nh_weight;
            }
        else if (!weightedHopPrefer
                 || pNxtHop->nh_weight == weight)
            {
            /* Select the LRU hop */
            BWMON_PRINTK ("Compare TS: (%d)%llu - (%d)%llu\n",
                          currHop, pNxtHop->nh_timestamp, selHop, timestamp);

            if (timestamp > pNxtHop->nh_timestamp)
                {
                selHop = currHop;
                timestamp = pNxtHop->nh_timestamp;
                }
            else if (timestamp == pNxtHop->nh_timestamp)
                {
                /* Select this hop, if the current hop is not the one that
                 * was previously the last time we came here
                 */
                if (currHop != lastHopSelected)
                    selHop = currHop;
                }
            }
        }

    pFib->nh_sel = selHop;

    /* update the timestamp of the selected hop */
    pNxtHop = &pFib->fi->fib_nh[selHop];
    pNxtHop->nh_timestamp = jiffies;

    BWMON_PRINTK ("%s-used hop %d(last:%d) selected as next hop for "
                  "SRC: "NIPQUAD_FMT "==> DST: " NIPQUAD_FMT" TS:%llu\n",
                  (totalBw == 0 || pNxtHop->linkUsage == NXTHOP_LINK_UNDERUSED)
                                                       ? "under" : "normal",
                  selHop, lastHopSelected, NIPQUAD(pFlow->fl4_src),
                  NIPQUAD(pFlow->fl4_dst), pNxtHop->nh_timestamp);

    lastHopSelected = selHop;

    return 0;
    }

#ifdef CONFIG_SYSCTL

/*******************************************************************************
*
* procMultipathSelect - handler for /proc/sys/net/bwmonitor/multipath_select
*
* This function installs/uninstalls the multipath selection hook.
*
* RETURNS: 0
*/
static int procMultipathSelect
    (
    ctl_table   * pTable,
    int           write,
    struct file * pFile,
    void __user * pBuffer,
    size_t      * pLen,
    loff_t      * pPos
    )
    {
    int result;

    result = proc_dointvec (pTable, write, pFile, pBuffer, pLen, pPos);

    if (write && result == 0)
        {
        if (multipathSelect)
            {
            /* Install the hook */
            bwmon_multipath_weight_check = bwMonMultipathNxtHopSelect;
            BWMON_PRINTK ("bwMonitor multipath selection enabled\n");
            }
        else
            {
            /* Uninstall the hook */
            bwmon_multipath_weight_check = NULL;
            BWMON_PRINTK ("bwMonitor multipath selection disabled\n");
            }
        }
    return result;
    }

static ctl_table bwMonitorSysctlTable[] =
    {
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "multipath_select",
        .mode     = 0644,
        .data     = &multipathSelect,
        .maxlen   = sizeof (int),
        .proc_handler = procMultipathSelect,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "weighted_hop_prefer",
        .mode     = 0644,
        .data     = &weightedHopPrefer,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "load_tolerance",
        .mode     = 0644,
        .data     = &loadTolerance,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
#ifdef BWMON_DEBUG
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "debug",
        .mode     = 0644,
        .data     = &debug,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "rate_interval",
        .mode     = 0644,
        .data     = &bwMonRateInterval,
        .maxlen   = sizeof (int),
        .proc_handler = proc_dointvec,
        },
#endif /* BWMON_DEBUG */
        { 0 }
    };

static ctl_table bwMonitorNetTable[] =
    {
        {
        .ctl_name = CTL_UNNUMBERED,
        .procname = "bwmonitor",
        .mode = 0555,
        .child = bwMonitorSysctlTable
        },
        { 0 }
    };

static ctl_table bwMonitorRootTable[] =
    {
        {
        .ctl_name = CTL_NET,
        .procname = "net",
        .mode = 0555,
        .child = bwMonitorNetTable
        },
        { 0 }
    };

/*******************************************************************************
*
* bwMonitorSysctlRegister - registers sysctl entries
*
* This routine registers net.bwmonitor sysctl entries.
*
* RETURNS: 0 or -1
*/
static __init int bwMonitorSysctlRegister (void)
    {
    pBwMonitorSysctlHeader = register_sysctl_table (bwMonitorRootTable);
    if (pBwMonitorSysctlHeader == NULL)
        {
        BWMON_PRINTK ("bwMonitor sysctl registration failed\n");
        return -1;
        }

    BWMON_PRINTK ("bwMonitor sysctl registered\n");
    return 0;
    }

/*******************************************************************************
*
* bwMonitorSysctlRegister - registers sysctl entries
*
* This routine unregisters net.bwmonitor sysctl entries.
*
* RETURNS: 0 or -1
*/
static __exit int bwMonitorSysctlUnregister (void)
    {
    if (pBwMonitorSysctlHeader == NULL)
        {
        BWMON_PRINTK ("bwMonitor sysctl not registered\n");
        return -1;
        }

    unregister_sysctl_table (pBwMonitorSysctlHeader);
    pBwMonitorSysctlHeader = NULL;

    BWMON_PRINTK ("Unregistered bwMonitor sysctl\n");
    return 0;
    }
#endif /* CONFIG_SYSCTL */

/*******************************************************************************
*
* bwMonitorInit - initializes bandwidth monitor
*
* This routine initializes bandwidth monitor.
*
* RETURNS: 0 or -1
*/
int __init bwMonitorInit (void)
    {
#ifdef BWMON_DEBUG
    /* initialize the timer to calculate traffic rate */
	init_timer(&bwRateTimer);
	bwRateTimer.data = 0;
	bwRateTimer.function = bwMonTimerHandler;
	bwRateTimer.expires = jiffies + bwMonRateInterval * HZ;
	add_timer(&bwRateTimer);
#endif /* BWMON_DEBUG */

    bwmon_multipath_weight_check = bwMonMultipathNxtHopSelect;

#ifdef CONFIG_SYSCTL
    bwMonitorSysctlRegister();
#endif /* CONFIG_SYSCTL */
    BWMON_PRINTK ("Initialized bandwidth monitor ...\n");

    return (0);
    }

/*******************************************************************************
*
* bwMonitorExit - de-initalize the bandwidth monitor
*
* This routine deinitializes the bandwidth monitor.
*
* RETURNS: NONE
*/
void __exit bwMonitorExit (void)
    {
#ifdef BWMON_DEBUG
    /* de-initialize the timer traffic rate timer */
	del_timer(&bwRateTimer);
#endif /* BWMON_DEBUG */

#ifdef CONFIG_SYSCTL
    bwMonitorSysctlUnregister();
#endif /* CONFIG_SYSCTL */
    bwmon_multipath_weight_check = NULL;
    }

MODULE_AUTHOR("Samba <samba@teamf1.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("rate based multipath selector");
module_init(bwMonitorInit);
module_exit(bwMonitorExit);
